package net.sf.snmpadaptor4j;

import java.net.URL;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import javax.management.MBeanAttributeInfo;
import javax.management.MBeanConstructorInfo;
import javax.management.MBeanInfo;
import javax.management.MBeanNotificationInfo;
import javax.management.MBeanOperationInfo;
import javax.management.MBeanParameterInfo;
import javax.management.MBeanServer;
import javax.management.NotCompliantMBeanException;
import javax.management.ObjectName;
import javax.management.StandardMBean;
import net.sf.snmpadaptor4j.daemon.SnmpDaemon;
import net.sf.snmpadaptor4j.daemon.conf.SnmpAppContext;
import net.sf.snmpadaptor4j.daemon.conf.SnmpConfiguration;
import net.sf.snmpadaptor4j.daemon.conf.xml.XmlConfigParser;
import net.sf.snmpadaptor4j.daemon.mib.jmx.JmxSnmpMib;
import net.sf.snmpadaptor4j.daemon.mib.jmx.JvmSnmpMib;
import net.sf.snmpadaptor4j.daemon.mib.system.SystemSnmpMib;
import net.sf.snmpadaptor4j.daemon.opennms.OpennmsSnmpDaemon;
import net.sf.snmpadaptor4j.mbean.SystemInfo;
import org.apache.log4j.Logger;

/**
 * Implements a JMX adaptor for SNMP protocol.
 * @author Jean-Philippe MINETTI
 */
public class SnmpAdaptor
		extends StandardMBean
		implements SnmpConfiguration, SnmpAppContext, SnmpAdaptorMBean {

	/**
	 * Logger.
	 */
	private final Logger logger = Logger.getLogger(SnmpAdaptor.class);

	/**
	 * SNMP daemon.
	 */
	private final SnmpDaemon daemon;

	/**
	 * Informations on the system.
	 */
	private final SystemInfo systemInfo;

	/**
	 * Listening IP address of SNMP daemon (127.0.0.1 by default).
	 */
	private String listenerAddress = null;

	/**
	 * UDP port of SNMP daemon (161 by default).
	 */
	private Integer listenerPort = null;

	/**
	 * Protocol version of SNMP daemon (SNMP v2 by default).
	 */
	private Integer listenerSnmpVersion = null;

	/**
	 * Read community of SNMP daemon ("public" by default).
	 */
	private String listenerReadCommunity = null;

	/**
	 * Write community of SNMP daemon ("private" by default).
	 */
	private String listenerWriteCommunity = null;

	/**
	 * Default root OID containing the attributes of the application.
	 */
	private final String defaultRootOid;

	/**
	 * Map of root OIDs where the attributes of the application will stay.
	 */
	private final Map<String, String> rootOidMap;

	/**
	 * Map of MBean OIDs.
	 */
	private final Map<ObjectName, String> mBeanOidMap;

	/**
	 * Application context map.
	 */
	private final Map<ClassLoader, SnmpAppContext> appContextMap = new HashMap<ClassLoader, SnmpAppContext>();

	/**
	 * <b>M</b>anagement <b>I</b>nformation <b>B</b>ase (MIB) for system attributes.
	 */
	private final SystemSnmpMib systemMib;

	/**
	 * <b>M</b>anagement <b>I</b>nformation <b>B</b>ase (MIB) for attributes of JVM.
	 */
	private final JvmSnmpMib jvmMib;

	/**
	 * <b>M</b>anagement <b>I</b>nformation <b>B</b>ase (MIB) for access to JMX.
	 */
	private final JmxSnmpMib jmxMib;

	/**
	 * Constructor.
	 * <p>
	 * Set <code>TRUE</code> to <code>classLoaderScope</code> for web applications: this avoids to publish MBeans of other applications of application server.
	 * </p>
	 * @param classLoaderScope <code>TRUE</code> for handle only MBeans created by the same {@link ClassLoader} that the SNMP adapter. <code>FALSE</code> for handle
	 *            all MBeans of the JVM.
	 * @throws NotCompliantMBeanException if the <var>mbeanInterface</var> does not follow JMX design patterns for Management Interfaces, or if <var>this</var> does
	 *             not implement the specified interface.
	 */
	public SnmpAdaptor (final boolean classLoaderScope) throws NotCompliantMBeanException {
		this(null, null, null, classLoaderScope);
	}

	/**
	 * Constructor.
	 * <p>
	 * Set <code>TRUE</code> to <code>classLoaderScope</code> for web applications: this avoids to publish MBeans of other applications of application server.
	 * </p>
	 * @param appContext Application context.
	 * @param classLoaderScope <code>TRUE</code> for handle only MBeans created by the same {@link ClassLoader} that the SNMP adapter. <code>FALSE</code> for handle
	 *            all MBeans of the JVM.
	 * @throws NotCompliantMBeanException if the <var>mbeanInterface</var> does not follow JMX design patterns for Management Interfaces, or if <var>this</var> does
	 *             not implement the specified interface.
	 */
	public SnmpAdaptor (final SnmpAppContext appContext, final boolean classLoaderScope) throws NotCompliantMBeanException {
		this(null, appContext, null, classLoaderScope);
	}

	/**
	 * Constructor.
	 * <p>
	 * Set <code>TRUE</code> to <code>classLoaderScope</code> for web applications: this avoids to publish MBeans of other applications of application server.
	 * </p>
	 * @param url {@link URL} of SNMP configuration file (XML).
	 * @param classLoaderScope <code>TRUE</code> for handle only MBeans created by the same {@link ClassLoader} that the SNMP adapter. <code>FALSE</code> for handle
	 *            all MBeans of the JVM.
	 * @throws Exception Exception if an error occurred.
	 */
	public SnmpAdaptor (final URL url, final boolean classLoaderScope) throws Exception {
		this(url, null, classLoaderScope);
	}

	/**
	 * Constructor.
	 * <p>
	 * Set <code>TRUE</code> to <code>classLoaderScope</code> for web applications: this avoids to publish MBeans of other applications of application server.
	 * </p>
	 * @param url {@link URL} of SNMP configuration file (XML).
	 * @param systemInfo Informations on the system.
	 * @param classLoaderScope <code>TRUE</code> for handle only MBeans created by the same {@link ClassLoader} that the SNMP adapter. <code>FALSE</code> for handle
	 *            all MBeans of the JVM.
	 * @throws Exception Exception if an error occurred.
	 */
	public SnmpAdaptor (final URL url, final SystemInfo systemInfo, final boolean classLoaderScope) throws Exception {
		this(XmlConfigParser.newInstance(url), systemInfo, classLoaderScope);
	}

	/**
	 * Hidden constructor.
	 * <p>
	 * Set <code>TRUE</code> to <code>classLoaderScope</code> for web applications: this avoids to publish MBeans of other applications of application server.
	 * </p>
	 * @param xmlConfigParser Parser of SNMP configuration file (XML).
	 * @param systemInfo Informations on the system.
	 * @param classLoaderScope <code>TRUE</code> for handle only MBeans created by the same {@link ClassLoader} that the SNMP adapter. <code>FALSE</code> for handle
	 *            all MBeans of the JVM.
	 * @throws NotCompliantMBeanException if the <var>mbeanInterface</var> does not follow JMX design patterns for Management Interfaces, or if <var>this</var> does
	 *             not implement the specified interface.
	 */
	private SnmpAdaptor (final XmlConfigParser xmlConfigParser, final SystemInfo systemInfo, final boolean classLoaderScope) throws NotCompliantMBeanException {
		this(xmlConfigParser, xmlConfigParser, systemInfo, classLoaderScope);
	}

	/**
	 * Constructor.
	 * <p>
	 * Set <code>TRUE</code> to <code>classLoaderScope</code> for web applications: this avoids to publish MBeans of other applications of application server.
	 * </p>
	 * @param configuration SNMP configuration settings.
	 * @param appContext Application context.
	 * @param systemInfo Informations on the system.
	 * @param classLoaderScope <code>TRUE</code> for handle only MBeans created by the same {@link ClassLoader} that the SNMP adapter. <code>FALSE</code> for handle
	 *            all MBeans of the JVM.
	 * @throws NotCompliantMBeanException if the <var>mbeanInterface</var> does not follow JMX design patterns for Management Interfaces, or if <var>this</var> does
	 *             not implement the specified interface.
	 */
	public SnmpAdaptor (final SnmpConfiguration configuration, final SnmpAppContext appContext, final SystemInfo systemInfo, final boolean classLoaderScope)
			throws NotCompliantMBeanException {
		super(SnmpAdaptorMBean.class);

		// Descriptions
		final String classDescription = "JMX adaptor for SNMP protocol";
		final String listenerAddressAttributeDescription = "Listening IP address of SNMP daemon";
		final String listenerPortAttributeDescription = "Listening UDP port of SNMP daemon";
		final String listenerSnmpVersionAttributeDescription = "Protocol version of SNMP daemon (1 or 2)";
		final String listenerReadCommunityAttributeDescription = "Read community of SNMP daemon";
		final String listenerWriteCommunityAttributeDescription = "Write community of SNMP daemon";
		final String startedAttributeDescription = "TRUE if the SNMP daemon is started";
		final String addAppContextDescription = "Adds a new application context";
		final String removeAppContextDescription = "Removes an application context";
		final String startDescription = "Starts the SNMP daemon";
		final String stopDescription = "Stops the SNMP daemon";
		final String classLoaderParameterDescription = "Class loader of application";
		final String appContextParameterDescription = "Application context (any class that implements the interface " + SnmpAppContext.class.getName() + ")";
		final String urlParameterDescription = "URL of SNMP configuration file (XML)";
		final String configurationParameterDescription = "SNMP configuration settings (any class that implements the interface " + SnmpConfiguration.class.getName()
				+ ")";
		final String systemInfoParameterDescription = "Informations on the Java application";
		final String classLoaderScopeParameterDescription = "TRUE for handle only MBeans created by the same ClassLoader that the SNMP adapter - FALSE for handle all MBeans of the JVM (set TRUE to classLoaderScope for web applications: this avoids to publish MBeans of other applications of application server)";

		// Parameter info
		final MBeanParameterInfo classLoaderParameter = new MBeanParameterInfo("classLoader", ClassLoader.class.getName(), classLoaderParameterDescription);

		final MBeanParameterInfo appContextParameter = new MBeanParameterInfo("appContext", SnmpAppContext.class.getName(), appContextParameterDescription);
		final MBeanParameterInfo urlParameter = new MBeanParameterInfo("url", URL.class.getName(), urlParameterDescription);
		final MBeanParameterInfo configurationParameter = new MBeanParameterInfo("configuration", SnmpConfiguration.class.getName(),
				configurationParameterDescription);
		final MBeanParameterInfo systemInfoParameter = new MBeanParameterInfo("systemInfo", SystemInfo.class.getName(), systemInfoParameterDescription);
		final MBeanParameterInfo classLoaderScopeParameter = new MBeanParameterInfo("classLoaderScope", boolean.class.getName(),
				classLoaderScopeParameterDescription);

		// Operation info
		final MBeanOperationInfo addAppContextOperation = new MBeanOperationInfo("addAppContext", addAppContextDescription, new MBeanParameterInfo[] {
				classLoaderParameter, appContextParameter }, void.class.getName(), MBeanOperationInfo.ACTION);
		final MBeanOperationInfo removeAppContextOperation = new MBeanOperationInfo("removeAppContext", removeAppContextDescription,
				new MBeanParameterInfo[] { classLoaderParameter }, void.class.getName(), MBeanOperationInfo.ACTION);
		final MBeanOperationInfo startOperation = new MBeanOperationInfo("start", startDescription, new MBeanParameterInfo[] {}, void.class.getName(),
				MBeanOperationInfo.ACTION);
		final MBeanOperationInfo stopOperation = new MBeanOperationInfo("stop", stopDescription, new MBeanParameterInfo[] {}, void.class.getName(),
				MBeanOperationInfo.ACTION);
		final MBeanOperationInfo[] operations = new MBeanOperationInfo[] { addAppContextOperation, removeAppContextOperation, startOperation, stopOperation };

		// Constructor info
		final MBeanConstructorInfo constructor1 = new MBeanConstructorInfo(getClass().getName(), "Constructor with only classLoaderScope parameter",
				new MBeanParameterInfo[] { classLoaderScopeParameter });
		final MBeanConstructorInfo constructor2 = new MBeanConstructorInfo(getClass().getName(), "Constructor with only appContext and classLoaderScope parameters",
				new MBeanParameterInfo[] { appContextParameter, classLoaderScopeParameter });
		final MBeanConstructorInfo constructor3 = new MBeanConstructorInfo(getClass().getName(), "Constructor with only url and classLoaderScope parameters",
				new MBeanParameterInfo[] { urlParameter, classLoaderScopeParameter });
		final MBeanConstructorInfo constructor4 = new MBeanConstructorInfo(getClass().getName(),
				"Constructor with only url, systemInfo and classLoaderScope parameters", new MBeanParameterInfo[] { urlParameter, systemInfoParameter,
						classLoaderScopeParameter });
		final MBeanConstructorInfo constructor5 = new MBeanConstructorInfo(getClass().getName(),
				"Constructor with configuration, appContext, systemInfo and classLoaderScope parameters", new MBeanParameterInfo[] { configurationParameter,
						appContextParameter, systemInfoParameter, classLoaderScopeParameter });
		final MBeanConstructorInfo[] constructors = new MBeanConstructorInfo[] { constructor1, constructor2, constructor3, constructor4, constructor5 };

		// Attribute info
		final MBeanAttributeInfo listenerAddressAttribute = new MBeanAttributeInfo("ListenerAddress", String.class.getName(), listenerAddressAttributeDescription,
				true, true, false);
		final MBeanAttributeInfo listenerPortAttribute = new MBeanAttributeInfo("ListenerPort", Integer.class.getName(), listenerPortAttributeDescription, true,
				true, false);
		final MBeanAttributeInfo listenerSnmpVersionAttribute = new MBeanAttributeInfo("ListenerSnmpVersion", Integer.class.getName(),
				listenerSnmpVersionAttributeDescription, true, true, false);
		final MBeanAttributeInfo listenerReadCommunityAttribute = new MBeanAttributeInfo("ListenerReadCommunity", String.class.getName(),
				listenerReadCommunityAttributeDescription, true, true, false);
		final MBeanAttributeInfo listenerWriteCommunityAttribute = new MBeanAttributeInfo("ListenerWriteCommunity", String.class.getName(),
				listenerWriteCommunityAttributeDescription, true, true, false);
		final MBeanAttributeInfo startedAttribute = new MBeanAttributeInfo("Started", boolean.class.getName(), startedAttributeDescription, true, false, true);
		final MBeanAttributeInfo[] attributes = new MBeanAttributeInfo[] { listenerAddressAttribute, listenerPortAttribute, listenerSnmpVersionAttribute,
				listenerReadCommunityAttribute, listenerWriteCommunityAttribute, startedAttribute };

		// MBeanInfo
		final MBeanInfo mBeanInfo = new MBeanInfo(getClass().getName(), classDescription, attributes, constructors, operations, new MBeanNotificationInfo[] {},
				getMBeanInfo().getDescriptor());
		cacheMBeanInfo(mBeanInfo);

		this.defaultRootOid = ((appContext != null) && (appContext.getDefaultRootOid() != null) ? appContext.getDefaultRootOid() : "1.3.6.1.4.1.99.12.8.1");
		this.rootOidMap = Collections.unmodifiableMap((appContext != null) && (appContext.getRootOidMap() != null) ? appContext.getRootOidMap()
				: new HashMap<String, String>());
		this.mBeanOidMap = Collections.unmodifiableMap((appContext != null) && (appContext.getMBeanOidMap() != null) ? appContext.getMBeanOidMap()
				: new HashMap<ObjectName, String>());
		this.systemInfo = (systemInfo != null ? systemInfo : new SystemInfo());
		this.systemMib = new SystemSnmpMib(this, this.appContextMap, this.systemInfo);
		this.jvmMib = new JvmSnmpMib(this.systemMib);
		this.jmxMib = new JmxSnmpMib(this.jvmMib, this, this.appContextMap, classLoaderScope);
		this.daemon = new OpennmsSnmpDaemon(this, this.jmxMib);
		setListenerAddress(configuration != null ? configuration.getListenerAddress() : null);
		setListenerPort(configuration != null ? configuration.getListenerPort() : null);
		setListenerSnmpVersion(configuration != null ? configuration.getListenerSnmpVersion() : null);
		setListenerReadCommunity(configuration != null ? configuration.getListenerReadCommunity() : null);
		setListenerWriteCommunity(configuration != null ? configuration.getListenerWriteCommunity() : null);
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.SnmpAdaptorMBean#addAppContext(java.lang.ClassLoader, net.sf.snmpadaptor4j.daemon.conf.SnmpAppContext)
	 */
	public void addAppContext (final ClassLoader classLoader, final SnmpAppContext appContext) {
		synchronized (this.appContextMap) {
			this.appContextMap.put(classLoader, appContext);
			this.systemMib.initSysObjectIDSet();
		}
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.SnmpAdaptorMBean#removeAppContext(java.lang.ClassLoader)
	 */
	public void removeAppContext (final ClassLoader classLoader) {
		synchronized (this.appContextMap) {
			this.appContextMap.remove(classLoader);
			this.systemMib.initSysObjectIDSet();
		}
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.conf.SnmpConfiguration#getListenerAddress()
	 */
	public final String getListenerAddress () {
		return this.listenerAddress;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.SnmpAdaptorMBean#setListenerAddress(java.lang.String)
	 */
	public final void setListenerAddress (final String listenerAddress) {
		String newListenerAddress = formatAttribute(listenerAddress);
		if (newListenerAddress == null) {
			newListenerAddress = "localhost";
		}
		if (this.listenerAddress != null ? !this.listenerAddress.equals(newListenerAddress) : true) {
			try {
				this.daemon.stop();
				this.listenerAddress = newListenerAddress;
			}
			catch (final Throwable e) {
				this.logger.error(e);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.conf.SnmpConfiguration#getListenerPort()
	 */
	public final Integer getListenerPort () {
		return this.listenerPort;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.SnmpAdaptorMBean#setListenerPort(java.lang.Integer)
	 */
	public final void setListenerPort (final Integer listenerPort) {
		final Integer newListenerPort = (listenerPort != null ? listenerPort : new Integer(161));
		if (this.listenerPort != null ? !this.listenerPort.equals(newListenerPort) : true) {
			try {
				this.daemon.stop();
				this.listenerPort = newListenerPort;
			}
			catch (final Throwable e) {
				this.logger.error(e);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.conf.SnmpConfiguration#getListenerSnmpVersion()
	 */
	public final Integer getListenerSnmpVersion () {
		return this.listenerSnmpVersion;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.SnmpAdaptorMBean#setListenerSnmpVersion(java.lang.Integer)
	 */
	public final void setListenerSnmpVersion (final Integer listenerSnmpVersion) {
		Integer newListenerSnmpVersion;
		if ((listenerSnmpVersion == null) || (listenerSnmpVersion.intValue() < 1) || (listenerSnmpVersion.intValue() > 2)) {
			newListenerSnmpVersion = new Integer(2);
		}
		else {
			newListenerSnmpVersion = listenerSnmpVersion;
		}
		if (this.listenerSnmpVersion != null ? !this.listenerSnmpVersion.equals(newListenerSnmpVersion) : true) {
			try {
				this.daemon.stop();
				this.listenerSnmpVersion = newListenerSnmpVersion;
			}
			catch (final Throwable e) {
				this.logger.error(e);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.conf.SnmpConfiguration#getListenerReadCommunity()
	 */
	public final String getListenerReadCommunity () {
		return this.listenerReadCommunity;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.SnmpAdaptorMBean#setListenerReadCommunity(java.lang.String)
	 */
	public final void setListenerReadCommunity (final String listenerReadCommunity) {
		String newListenerReadCommunity = formatAttribute(listenerReadCommunity);
		if (newListenerReadCommunity == null) {
			newListenerReadCommunity = "public";
		}
		if (this.listenerReadCommunity != null ? !this.listenerReadCommunity.equals(newListenerReadCommunity) : true) {
			try {
				this.daemon.stop();
				this.listenerReadCommunity = newListenerReadCommunity;
			}
			catch (final Throwable e) {
				this.logger.error(e);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.conf.SnmpConfiguration#getListenerWriteCommunity()
	 */
	public final String getListenerWriteCommunity () {
		return this.listenerWriteCommunity;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.SnmpAdaptorMBean#setListenerWriteCommunity(java.lang.String)
	 */
	public final void setListenerWriteCommunity (final String listenerWriteCommunity) {
		String newListenerWriteCommunity = formatAttribute(listenerWriteCommunity);
		if (newListenerWriteCommunity == null) {
			newListenerWriteCommunity = "private";
		}
		if (this.listenerWriteCommunity != null ? !this.listenerWriteCommunity.equals(newListenerWriteCommunity) : true) {
			try {
				this.daemon.stop();
				this.listenerWriteCommunity = newListenerWriteCommunity;
			}
			catch (final Throwable e) {
				this.logger.error(e);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.conf.SnmpAppContext#getDefaultRootOid()
	 */
	public final String getDefaultRootOid () {
		return this.defaultRootOid;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.conf.SnmpAppContext#getRootOidMap()
	 */
	public final Map<String, String> getRootOidMap () {
		return this.rootOidMap;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.conf.SnmpAppContext#getMBeanOidMap()
	 */
	public final Map<ObjectName, String> getMBeanOidMap () {
		return this.mBeanOidMap;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.SnmpAdaptorMBean#start()
	 */
	public void start () throws Exception {
		this.daemon.start();
		if (!this.daemon.isStarted()) {
			throw new Exception("The SNMP daemon did not start");
		}
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.SnmpAdaptorMBean#stop()
	 */
	public void stop () throws Exception {
		this.daemon.stop();
		if (this.daemon.isStarted()) {
			throw new Exception("The SNMP daemon has not been stopped");
		}
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.SnmpAdaptorMBean#isStarted()
	 */
	public final boolean isStarted () {
		return this.daemon.isStarted();
	}

	/**
	 * Returns the informations on the system (MBean registrable).
	 * @return Informations on the system.
	 */
	public final SystemInfo getSystemInfo () {
		return this.systemInfo;
	}

	/**
	 * Formats an attribute value.
	 * @param value Value to format.
	 * @return Value formatted.
	 */
	protected final String formatAttribute (final String value) {
		String result = value;
		if (result != null) {
			result = result.trim();
			if (result.length() == 0) {
				result = null;
			}
		}
		return result;
	}

	/*
	 * (non-Javadoc)
	 * @see javax.management.StandardMBean#preRegister(javax.management.MBeanServer, javax.management.ObjectName)
	 */
	@Override
	public ObjectName preRegister (final MBeanServer server, final ObjectName name) throws Exception {
		final ObjectName result = super.preRegister(server, name);
		this.jmxMib.open(server);
		this.jvmMib.open(server);
		return result;
	}

	/*
	 * (non-Javadoc)
	 * @see javax.management.StandardMBean#preDeregister()
	 */
	@Override
	public void preDeregister () throws Exception {
		super.preDeregister();
		this.jmxMib.close();
		this.jvmMib.close();
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString () {
		return "SnmpAdaptor[" + this.listenerAddress + ":" + this.listenerPort + "]";
	}

}