package net.sf.snmpadaptor4j.daemon.conf.xml;

import java.net.URL;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import javax.management.ObjectName;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.Unmarshaller;
import net.sf.snmpadaptor4j.daemon.conf.SnmpAppContext;
import net.sf.snmpadaptor4j.daemon.conf.SnmpConfiguration;
import net.sf.snmpadaptor4j.daemon.conf.xml.jaxb.Config;
import net.sf.snmpadaptor4j.daemon.conf.xml.jaxb.Daemon;
import net.sf.snmpadaptor4j.daemon.conf.xml.jaxb.MBean;
import net.sf.snmpadaptor4j.daemon.conf.xml.jaxb.Root;
import net.sf.snmpadaptor4j.daemon.conf.xml.jaxb.Roots;

/**
 * Parser of SNMP configuration file (XML).
 * @author Jean-Philippe MINETTI
 */
public final class XmlConfigParser
		implements SnmpConfiguration, SnmpAppContext {

	/**
	 * XSD object representing the daemon configuration.
	 */
	private final Daemon daemon;

	/**
	 * XSD object containing all root OIDs.
	 */
	private final Roots roots;

	/**
	 * Map of root OIDs where the attributes of the application will stay.
	 */
	private final Map<String, String> rootOidMap;

	/**
	 * Map of MBean OIDs.
	 */
	private final Map<ObjectName, String> mBeanOidMap;

	/**
	 * Creates and returns a new instance of {@link XmlConfigParser}.
	 * @param url {@link URL} to the SNMP configuration file.
	 * @return New instance of {@link XmlConfigParser}.
	 * @throws Exception Exception if an error occurred.
	 */
	public static XmlConfigParser newInstance (final URL url) throws Exception {
		final JAXBContext jaxbContext = JAXBContext.newInstance(Config.class.getPackage().getName());
		final Unmarshaller unmarshaller = jaxbContext.createUnmarshaller();
		@SuppressWarnings("unchecked")
		final JAXBElement<Config> jaxbElement = (JAXBElement<Config>) unmarshaller.unmarshal(url);
		final Config config = jaxbElement.getValue();

		// rootOidMap
		final Map<String, String> rootOidMap = new HashMap<String, String>();
		if (config.getRoots() != null) {
			for (final Root root : config.getRoots().getRoot()) {
				rootOidMap.put(root.getId(), root.getOid());
			}
		}

		// mBeanOidMap
		final Map<ObjectName, String> mBeanOidMap = new HashMap<ObjectName, String>();
		if (config.getMbeans() != null) {
			String oid;
			for (final MBean mBean : config.getMbeans().getMbean()) {
				oid = null;
				if (mBean.getRoot() != null) {
					oid = rootOidMap.get(mBean.getRoot());
				}
				if (oid == null) {
					oid = config.getRoots().getDefault();
				}
				oid = oid + "." + mBean.getOid();
				mBeanOidMap.put(new ObjectName(mBean.getName()), oid);
			}
		}

		return new XmlConfigParser(config.getDaemon(), config.getRoots(), rootOidMap, mBeanOidMap);
	}

	/**
	 * Constructor.
	 * @param daemon XSD object representing the daemon configuration.
	 * @param roots XSD object containing all root OIDs.
	 * @param rootOidMap Map of root OIDs where the attributes of the application will stay.
	 * @param mBeanOidMap Map of MBean OIDs.
	 */
	private XmlConfigParser (final Daemon daemon, final Roots roots, final Map<String, String> rootOidMap, final Map<ObjectName, String> mBeanOidMap) {
		super();
		this.daemon = daemon;
		this.roots = roots;
		this.rootOidMap = Collections.unmodifiableMap(rootOidMap);
		this.mBeanOidMap = Collections.unmodifiableMap(mBeanOidMap);
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.conf.SnmpConfiguration#getListenerAddress()
	 */
	public String getListenerAddress () {
		return (this.daemon != null ? this.daemon.getAddress() : null);
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.conf.SnmpConfiguration#getListenerPort()
	 */
	public Integer getListenerPort () {
		return (this.daemon != null ? new Integer(this.daemon.getPort()) : null);
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.conf.SnmpConfiguration#getListenerSnmpVersion()
	 */
	public Integer getListenerSnmpVersion () {
		return (this.daemon != null ? new Integer(this.daemon.getVersion()) : null);
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.conf.SnmpConfiguration#getListenerReadCommunity()
	 */
	public String getListenerReadCommunity () {
		return (this.daemon != null ? this.daemon.getReadCommunity() : null);
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.conf.SnmpConfiguration#getListenerWriteCommunity()
	 */
	public String getListenerWriteCommunity () {
		return (this.daemon != null ? this.daemon.getWriteCommunity() : null);
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.conf.SnmpAppContext#getDefaultRootOid()
	 */
	public String getDefaultRootOid () {
		return (this.roots != null ? this.roots.getDefault() : null);
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.conf.SnmpAppContext#getRootOidMap()
	 */
	public Map<String, String> getRootOidMap () {
		return this.rootOidMap;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.conf.SnmpAppContext#getMBeanOidMap()
	 */
	public Map<ObjectName, String> getMBeanOidMap () {
		return this.mBeanOidMap;
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString () {
		return "XmlConfigParser[listenerAddress=" + getListenerAddress() + ", listenerPort=" + getListenerPort() + ", listenerSnmpVersion="
				+ getListenerSnmpVersion() + ", listenerReadCommunity=" + getListenerReadCommunity() + ", listenerWriteCommunity=" + getListenerWriteCommunity()
				+ ", defaultRootOid=" + getDefaultRootOid() + ", rootOidMap=" + this.rootOidMap + ", mBeanOidMap=" + this.mBeanOidMap + "]";
	}

}