package net.sf.snmpadaptor4j.daemon.mib.jmx;

import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.management.openmbean.CompositeData;
import net.sf.snmpadaptor4j.daemon.mib.SnmpDataType;
import net.sf.snmpadaptor4j.daemon.mib.SnmpMibExternalNode;
import net.sf.snmpadaptor4j.daemon.mib.SnmpOid;

/**
 * Object representing an external node of the SNMP <b>M</b>anagement <b>I</b>nformation <b>B</b>ase (MIB) for {@link CompositeData}.
 * @author Jean-Philippe MINETTI
 */
public final class CompositeDataSnmpExternalNode
		implements SnmpMibExternalNode {

	/**
	 * Object Identifier of MIB node.
	 */
	private final SnmpOid oid;

	/**
	 * JMX agent where the MBean was registered.
	 */
	private final MBeanServer server;

	/**
	 * MBean name.
	 */
	private final ObjectName mBeanName;

	/**
	 * Attribute name of {@link CompositeData}.
	 */
	private final String attributeName;

	/**
	 * Attribute key of value.
	 */
	private final String attributeKey;

	/**
	 * SNMP data type of MIB node.
	 */
	private final SnmpDataType snmpDataType;

	/**
	 * Data type of JMX attribute.
	 */
	private final Class<?> jmxDataType;

	/**
	 * Constructor.
	 * @param oid Object Identifier of MIB node.
	 * @param server JMX agent where the MBean was registered.
	 * @param mBeanName MBean name.
	 * @param attributeName Attribute name of {@link CompositeData}.
	 * @param attributeKey Attribute key of value.
	 * @param snmpDataType SNMP data type of MIB node.
	 * @param jmxDataType Data type of JMX attribute.
	 */
	public CompositeDataSnmpExternalNode (final SnmpOid oid, final MBeanServer server, final ObjectName mBeanName, final String attributeName,
			final String attributeKey, final SnmpDataType snmpDataType, final Class<?> jmxDataType) {
		super();
		this.oid = oid;
		this.server = server;
		this.mBeanName = mBeanName;
		this.attributeName = attributeName;
		this.attributeKey = attributeKey;
		this.snmpDataType = snmpDataType;
		this.jmxDataType = jmxDataType;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.mib.SnmpMibExternalNode#getOid()
	 */
	public SnmpOid getOid () {
		return this.oid;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.mib.SnmpMibExternalNode#getSnmpDataType()
	 */
	public SnmpDataType getSnmpDataType () {
		return this.snmpDataType;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.mib.SnmpMibExternalNode#getJmxDataType()
	 */
	public Class<?> getJmxDataType () {
		return this.jmxDataType;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.mib.SnmpMibExternalNode#getValue()
	 */
	public Object getValue () throws Exception {
		Object value = null;
		final CompositeData compositeData = (CompositeData) this.server.getAttribute(this.mBeanName, this.attributeName);
		if (compositeData != null) {
			value = compositeData.get(this.attributeKey);
		}
		return value;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.mib.SnmpMibExternalNode#setValue(java.lang.Object)
	 */
	public void setValue (final Object value) {
		// NOP
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.mib.SnmpMibExternalNode#isReadable()
	 */
	public boolean isReadable () {
		return true;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.mib.SnmpMibExternalNode#isWritable()
	 */
	public boolean isWritable () {
		return false;
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString () {
		Object value;
		try {
			value = getValue();
		}
		catch (final Throwable e) {
			value = "ERROR: " + (e.getMessage() != null ? e.getMessage() : e.getClass().getName());
		}
		return this.oid + ": (" + getSnmpDataType() + ") " + value;
	}

}