package net.sf.snmpadaptor4j.daemon.mib.jmx;

import java.math.BigInteger;
import java.net.InetAddress;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Map.Entry;
import javax.management.MBeanAttributeInfo;
import javax.management.MBeanServer;
import javax.management.MBeanServerDelegate;
import javax.management.MBeanServerNotification;
import javax.management.Notification;
import javax.management.NotificationListener;
import javax.management.ObjectName;
import javax.management.relation.MBeanServerNotificationFilter;
import net.sf.snmpadaptor4j.daemon.conf.SnmpAppContext;
import net.sf.snmpadaptor4j.daemon.mib.SnmpDataType;
import net.sf.snmpadaptor4j.daemon.mib.SnmpMib;
import net.sf.snmpadaptor4j.daemon.mib.SnmpMibExternalNode;
import net.sf.snmpadaptor4j.daemon.mib.SnmpOid;
import net.sf.snmpadaptor4j.daemon.mib.jmx.xml.XmlMappingParser;
import net.sf.snmpadaptor4j.daemon.mib.jmx.xml.jaxb.MBean;
import net.sf.snmpadaptor4j.daemon.mib.jmx.xml.jaxb.MBeanAttribute;
import net.sf.snmpadaptor4j.daemon.mib.jmx.xml.jaxb.MBeanAttributes;
import org.apache.log4j.Logger;

/**
 * Object representing the <b>M</b>anagement <b>I</b>nformation <b>B</b>ase (MIB) for access to JMX.
 * @author Jean-Philippe MINETTI
 */
public final class JmxSnmpMib
		implements SnmpMib, NotificationListener {

	/**
	 * Logger.
	 */
	private final Logger logger = Logger.getLogger(JmxSnmpMib.class);

	/**
	 * Other object representing the MIB.
	 */
	private final SnmpMib other;

	/**
	 * Context of main application.
	 */
	private final SnmpAppContext mainAppContext;

	/**
	 * Application context map.
	 */
	private final Map<ClassLoader, SnmpAppContext> appContextMap;

	/**
	 * <code>TRUE</code> for handle only MBeans created by the same {@link ClassLoader} that the SNMP adapter. <code>FALSE</code> for handle all MBeans of the JVM.
	 */
	private final boolean classLoaderScope;

	/**
	 * <b>M</b>anagement <b>I</b>nformation <b>B</b>ase (MIB).
	 */
	private final SortedMap<SnmpOid, SnmpMibExternalNode> mib = new TreeMap<SnmpOid, SnmpMibExternalNode>();

	/**
	 * JMX agent.
	 */
	private MBeanServer jmxServer;

	/**
	 * Constructor.
	 * @param other Other object representing the MIB.
	 * @param mainAppContext Context of main application (must not be <code>NULL</code>).
	 * @param appContextMap Application context map.
	 * @param classLoaderScope <code>TRUE</code> for handle only MBeans created by the same {@link ClassLoader} that the SNMP adapter. <code>FALSE</code> for handle
	 *            all MBeans of the JVM.
	 */
	public JmxSnmpMib (final SnmpMib other, final SnmpAppContext mainAppContext, final Map<ClassLoader, SnmpAppContext> appContextMap, final boolean classLoaderScope) {
		super();
		this.other = other;
		this.mainAppContext = mainAppContext;
		this.appContextMap = appContextMap;
		this.classLoaderScope = classLoaderScope;
	}

	/**
	 * Opens the connection with the JMX agent.
	 * @param server JMX agent.
	 * @throws Exception Exception if an error occurred.
	 */
	public synchronized void open (final MBeanServer server) throws Exception {
		this.logger.trace("JMX opening...");
		if (this.jmxServer != null) {
			throw new Exception("Already connected to a JMX agent");
		}

		// JMX notification registering
		final MBeanServerNotificationFilter filter = new MBeanServerNotificationFilter();
		filter.enableAllObjectNames();
		server.addNotificationListener(MBeanServerDelegate.DELEGATE_NAME, this, filter, server);

		// Existing MBean parsing
		final Set<ObjectName> mBeanNameList = server.queryNames(null, null);
		for (final ObjectName mBeanName : mBeanNameList) {
			register(server, mBeanName);
		}

		this.jmxServer = server;
		this.logger.trace("JMX opened");
	}

	/**
	 * Closes the connection with the JMX agent.
	 * @throws Exception Exception if an error occurred.
	 */
	public synchronized void close () throws Exception {
		this.logger.trace("JMX closing...");
		if (this.jmxServer == null) {
			throw new Exception("Not connected to a JMX agent");
		}

		// JMX notification deregistering
		this.jmxServer.removeNotificationListener(MBeanServerDelegate.DELEGATE_NAME, this);

		// MIB cleaning
		JmxSnmpMibExternalNode node;
		synchronized (this.mib) {
			while (!this.mib.isEmpty()) {
				node = (JmxSnmpMibExternalNode) this.mib.get(this.mib.firstKey());
				unregister(node.getMBeanName(), node.getOid());
			}
		}

		this.jmxServer = null;
		this.logger.trace("JMX closed");
	}

	/*
	 * (non-Javadoc)
	 * @see javax.management.NotificationListener#handleNotification(javax.management.Notification, java.lang.Object)
	 */
	public synchronized void handleNotification (final Notification notification, final Object handback) {
		if ((handback instanceof MBeanServer) && (this.jmxServer == (MBeanServer) handback)) {
			if (notification instanceof MBeanServerNotification) {
				final MBeanServerNotification serverNotification = (MBeanServerNotification) notification;
				if (MBeanServerNotification.REGISTRATION_NOTIFICATION.equals(serverNotification.getType())) {
					try {
						register((MBeanServer) handback, serverNotification.getMBeanName());
					}
					catch (final Throwable e) {
						this.logger.error(serverNotification.getMBeanName() + ": MBean not loaded in SNMP adapter", e);
					}
				}
				else if (MBeanServerNotification.UNREGISTRATION_NOTIFICATION.equals(serverNotification.getType())) {
					try {
						unregister(serverNotification.getMBeanName());
					}
					catch (final Throwable e) {
						this.logger.error(serverNotification.getMBeanName() + ": MBean not unloaded in SNMP adapter", e);
					}
				}
			}
		}
	}

	/**
	 * Registers a MBean to SNMP <b>M</b>anagement <b>I</b>nformation <b>B</b>ase (MIB).
	 * @param server JMX agent.
	 * @param mBeanName MBean name.
	 * @throws Exception Exception if an error occurred.
	 */
	private void register (final MBeanServer server, final ObjectName mBeanName) throws Exception {
		if (server.isRegistered(mBeanName)) {
			final ClassLoader classLoader = server.getClassLoaderFor(mBeanName);
			if (!this.classLoaderScope || this.getClass().getClassLoader().equals(classLoader)) {
				final Class<?> mBeanClass = classLoader.loadClass(server.getObjectInstance(mBeanName).getClassName());
				final URL url = mBeanClass.getResource(mBeanClass.getSimpleName() + ".snmp.xml");
				if (url != null) {
					if (this.logger.isDebugEnabled()) {
						this.logger.debug("SNMP mapping found at " + url);
					}
					final XmlMappingParser parser = new XmlMappingParser(url);
					parser.load();
					final String baseOid = findOid(classLoader, mBeanName, parser);
					if (baseOid != null) {
						register(server, mBeanName, classLoader, baseOid, parser.getMapping().getAttributes());
					}
					else if (this.logger.isDebugEnabled()) {
						this.logger.debug("None OID found for [" + mBeanName + "]");
					}
				}
				else if (this.logger.isDebugEnabled()) {
					this.logger.debug("SNMP mapping missing for [" + mBeanName + "]");
				}
			}
			else if (this.logger.isTraceEnabled()) {
				this.logger.trace("SNMP mapping ignored because classLoaderScope = TRUE for [" + mBeanName + "]");
			}
		}
	}

	/**
	 * Registers a MBean to SNMP <b>M</b>anagement <b>I</b>nformation <b>B</b>ase (MIB).
	 * @param server JMX agent.
	 * @param mBeanName MBean name.
	 * @param classLoader {@link ClassLoader} where the MBean has been created.
	 * @param baseOid Base OID of MBean instance.
	 * @param attributes Attributes settings from mapping file.
	 * @throws Exception Exception if an error occurred.
	 */
	private void register (final MBeanServer server, final ObjectName mBeanName, final ClassLoader classLoader, final String baseOid,
			final MBeanAttributes attributes) throws Exception {
		final Map<String, MBeanAttributeInfo> attributeMap = new HashMap<String, MBeanAttributeInfo>();
		for (final MBeanAttributeInfo info : server.getMBeanInfo(mBeanName).getAttributes()) {
			attributeMap.put(info.getName(), info);
		}
		final Iterator<MBeanAttribute> attributeIterator = attributes.getAttribute().iterator();
		MBeanAttribute attribute;
		String attributeName;
		MBeanAttributeInfo attributeInfo;
		SnmpOid oid;
		SnmpDataType snmpDataType;
		Class<?> jmxDataType;
		boolean readable;
		boolean writable;
		while (attributeIterator.hasNext()) {
			attribute = attributeIterator.next();

			// Property name & info
			attributeName = attribute.getName().substring(0, 1).toUpperCase() + attribute.getName().substring(1);
			attributeInfo = attributeMap.get(attributeName);

			// OID
			oid = SnmpOid.newInstance(baseOid, attribute.getNode(), 0);

			// snmpDataType
			snmpDataType = SnmpDataType.valueOf(attribute.getType());

			// jmxDataType
			jmxDataType = toClass(classLoader, attributeInfo.getType());

			// readable
			readable = (attributeInfo.isReadable() && !attribute.isDisabled());

			// writable
			writable = (readable && attributeInfo.isWritable() && attribute.isWritable());

			if (this.logger.isInfoEnabled()) {
				this.logger.info("SNMP attribute registering at " + oid + " (" + snmpDataType + ") = [" + mBeanName + "]." + attribute.getName() + " ("
						+ jmxDataType.getSimpleName() + ")");
			}
			checkDataType(snmpDataType, jmxDataType);
			synchronized (this.mib) {
				this.mib.put(oid, new JmxSnmpMibExternalNode(oid, server, mBeanName, attributeName, snmpDataType, jmxDataType, readable, writable));
			}
		}
	}

	/**
	 * Unregisters a MBean of SNMP <b>M</b>anagement <b>I</b>nformation <b>B</b>ase (MIB).
	 * @param mBeanName MBean name.
	 */
	private void unregister (final ObjectName mBeanName) {
		final Set<SnmpOid> oidSet = new TreeSet<SnmpOid>();
		synchronized (this.mib) {
			for (final SnmpMibExternalNode node : this.mib.values()) {
				if (mBeanName.equals(((JmxSnmpMibExternalNode) node).getMBeanName())) {
					oidSet.add(node.getOid());
				}
			}
			for (final SnmpOid oid : oidSet) {
				unregister(mBeanName, oid);
			}
		}
	}

	/**
	 * Unregisters a MBean of SNMP <b>M</b>anagement <b>I</b>nformation <b>B</b>ase (MIB).
	 * @param mBeanName MBean name.
	 * @param oid OID of parameter to unregister.
	 */
	private void unregister (final ObjectName mBeanName, final SnmpOid oid) {
		if (this.logger.isInfoEnabled()) {
			this.logger.info("SNMP attribute unregistering at " + oid + " for [" + mBeanName + "]");
		}
		this.mib.remove(oid);
	}

	/**
	 * Finds a base OID of a MBean instance.
	 * @param classLoader {@link ClassLoader} where the MBean has been created.
	 * @param mBeanName MBean name.
	 * @param parser Parser of SNMP mapping file (XML).
	 * @return Base OID found.
	 */
	private String findOid (final ClassLoader classLoader, final ObjectName mBeanName, final XmlMappingParser parser) {

		// Application context finding
		SnmpAppContext ctx = null;
		ClassLoader cl = classLoader;
		while ((ctx == null) && (cl != null)) {
			ctx = this.appContextMap.get(cl);
			cl = cl.getParent();
		}
		if (ctx == null) {
			ctx = this.mainAppContext;
		}

		// OID finding
		String baseOid = (ctx.getMBeanOidMap() != null ? ctx.getMBeanOidMap().get(mBeanName) : null);
		if ((baseOid == null) && (parser.getMapping().getMbeans() != null)) {
			final Iterator<MBean> mBeanIterator = parser.getMapping().getMbeans().getMbean().iterator();
			final String desiredName = mBeanName.toString();
			MBean mBean;
			while (mBeanIterator.hasNext() && (baseOid == null)) {
				mBean = mBeanIterator.next();
				if (mBean.getName().equals(desiredName)) {
					if (mBean.getRoot() != null) {
						baseOid = ctx.getRootOidMap().get(mBean.getRoot());
					}
					if (baseOid == null) {
						baseOid = ctx.getDefaultRootOid();
					}
					if (baseOid == null) {
						baseOid = this.mainAppContext.getDefaultRootOid();
					}
					baseOid = baseOid + "." + mBean.getOid();
				}
			}
		}

		return baseOid;
	}

	/**
	 * Converts a class name to its instance.
	 * @param classLoader {@link ClassLoader} where the MBean has been created.
	 * @param className Class name.
	 * @return Class instance.
	 * @throws Exception Exception if an error occurred.
	 */
	private Class<?> toClass (final ClassLoader classLoader, final String className) throws Exception {
		Class<?> result;
		if (byte.class.getName().equals(className)) {
			result = byte.class;
		}
		else if (short.class.getName().equals(className)) {
			result = short.class;
		}
		else if (int.class.getName().equals(className)) {
			result = int.class;
		}
		else if (long.class.getName().equals(className)) {
			result = long.class;
		}
		else if (byte[].class.getName().equals(className)) {
			result = byte[].class;
		}
		else if (short[].class.getName().equals(className)) {
			result = short[].class;
		}
		else if (int[].class.getName().equals(className)) {
			result = int[].class;
		}
		else if (long[].class.getName().equals(className)) {
			result = long[].class;
		}
		else {
			result = classLoader.loadClass(className);
		}
		return result;
	}

	/**
	 * Checks the consistency between the data types SNMP and Java.
	 * @param snmpDataType SNMP data type of MIB node.
	 * @param jmxDataType Data type of JMX attribute.
	 * @throws Exception Exception if an error occurred.
	 */
	private void checkDataType (final SnmpDataType snmpDataType, final Class<?> jmxDataType) throws Exception {
		if (snmpDataType == SnmpDataType.integer32) {
			if (!int.class.equals(jmxDataType) && !Integer.class.equals(jmxDataType) && !byte.class.equals(jmxDataType) && !Byte.class.equals(jmxDataType)
					&& !short.class.equals(jmxDataType) && !Short.class.equals(jmxDataType)) {
				throw new Exception(snmpDataType + " is inconsistent with " + jmxDataType.getName());
			}
		}
		else if ((snmpDataType == SnmpDataType.unsigned32) || (snmpDataType == SnmpDataType.gauge32) || (snmpDataType == SnmpDataType.counter32)) {
			if (!long.class.equals(jmxDataType) && !Long.class.equals(jmxDataType) && !int.class.equals(jmxDataType) && !Integer.class.equals(jmxDataType)
					&& !byte.class.equals(jmxDataType) && !Byte.class.equals(jmxDataType) && !short.class.equals(jmxDataType) && !Short.class.equals(jmxDataType)) {
				throw new Exception(snmpDataType + " is inconsistent with " + jmxDataType.getName());
			}
		}
		else if (snmpDataType == SnmpDataType.counter64) {
			if (!BigInteger.class.equals(jmxDataType) && !long.class.equals(jmxDataType) && !Long.class.equals(jmxDataType)) {
				throw new Exception(snmpDataType + " is inconsistent with " + jmxDataType.getName());
			}
		}
		else if (snmpDataType == SnmpDataType.timeTicks) {
			if (!long.class.equals(jmxDataType) && !Long.class.equals(jmxDataType)) {
				throw new Exception(snmpDataType + " is inconsistent with " + jmxDataType.getName());
			}
		}
		else if (snmpDataType == SnmpDataType.ipAddress) {
			if (!String.class.equals(jmxDataType) && !InetAddress.class.equals(jmxDataType)) {
				throw new Exception(snmpDataType + " is inconsistent with " + jmxDataType.getName());
			}
		}
		else if (snmpDataType == SnmpDataType.objectIdentifier) {
			if (!String.class.equals(jmxDataType) && !SnmpOid.class.equals(jmxDataType)) {
				throw new Exception(snmpDataType + " is inconsistent with " + jmxDataType.getName());
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.mib.SnmpMib#find(net.sf.snmpadaptor4j.daemon.mib.SnmpOid)
	 */
	public SnmpMibExternalNode find (final SnmpOid oid) {
		SnmpMibExternalNode node;
		synchronized (this.mib) {
			node = this.mib.get(oid);
		}
		if (node == null) {
			node = this.other.find(oid);
		}
		return node;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.mib.SnmpMib#next(net.sf.snmpadaptor4j.daemon.mib.SnmpOid)
	 */
	public SnmpMibExternalNode next (final SnmpOid oid) {
		SnmpMibExternalNode node;
		final Iterator<Entry<SnmpOid, SnmpMibExternalNode>> entryIterator = nextSet(oid).entrySet().iterator();
		if (entryIterator.hasNext()) {
			node = entryIterator.next().getValue();
		}
		else {
			node = null;
		}
		return node;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.mib.SnmpMib#nextSet(net.sf.snmpadaptor4j.daemon.mib.SnmpOid)
	 */
	public SortedMap<SnmpOid, SnmpMibExternalNode> nextSet (final SnmpOid oid) {
		final SortedMap<SnmpOid, SnmpMibExternalNode> nodeMap = new TreeMap<SnmpOid, SnmpMibExternalNode>();
		synchronized (this.mib) {
			nodeMap.putAll(this.mib.tailMap(SnmpOid.newInstance(oid.getOid(), 0)));
		}
		nodeMap.putAll(this.other.nextSet(oid));
		return nodeMap;
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString () {
		return "JmxSnmpMib[" + this.mib.values() + "]";
	}

}