package net.sf.snmpadaptor4j.daemon.mib.jmx;

import javax.management.Attribute;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import net.sf.snmpadaptor4j.daemon.mib.SnmpDataType;
import net.sf.snmpadaptor4j.daemon.mib.SnmpMibExternalNode;
import net.sf.snmpadaptor4j.daemon.mib.SnmpOid;

/**
 * Object representing an external node of the SNMP <b>M</b>anagement <b>I</b>nformation <b>B</b>ase (MIB) and enabling to access an MBean attribute
 * (reading/writing).
 * @author Jean-Philippe MINETTI
 */
public final class JmxSnmpMibExternalNode
		implements SnmpMibExternalNode {

	/**
	 * Object Identifier of MIB node.
	 */
	private final SnmpOid oid;

	/**
	 * JMX agent where the MBean was registered.
	 */
	private final MBeanServer server;

	/**
	 * MBean name.
	 */
	private final ObjectName mBeanName;

	/**
	 * Attribute name.
	 */
	private final String attributeName;

	/**
	 * SNMP data type of MIB node.
	 */
	private final SnmpDataType snmpDataType;

	/**
	 * Data type of JMX attribute.
	 */
	private final Class<?> jmxDataType;

	/**
	 * <code>TRUE</code> if the attribute can be read (for SNMP write and read community), <code>FALSE</code> otherwise.
	 */
	private final boolean readable;

	/**
	 * <code>TRUE</code> if the attribute can be write (for SNMP write community), <code>FALSE</code> otherwise.
	 */
	private final boolean writable;

	/**
	 * Constructor.
	 * @param oid Object Identifier of MIB node.
	 * @param server JMX agent where the MBean was registered.
	 * @param mBeanName MBean name.
	 * @param attributeName Attribute name.
	 * @param snmpDataType SNMP data type of MIB node.
	 * @param jmxDataType Data type of JMX attribute.
	 * @param readable <code>TRUE</code> if the attribute can be read (for SNMP write and read community), <code>FALSE</code> otherwise.
	 * @param writable <code>TRUE</code> if the attribute can be write (for SNMP write community), <code>FALSE</code> otherwise.
	 */
	protected JmxSnmpMibExternalNode (final SnmpOid oid, final MBeanServer server, final ObjectName mBeanName, final String attributeName,
			final SnmpDataType snmpDataType, final Class<?> jmxDataType, final boolean readable, final boolean writable) {
		super();
		this.oid = oid;
		this.server = server;
		this.mBeanName = mBeanName;
		this.attributeName = attributeName;
		this.snmpDataType = snmpDataType;
		this.jmxDataType = jmxDataType;
		this.readable = readable;
		this.writable = writable;
	}

	/**
	 * Returns the MBean name.
	 * @return MBean name.
	 */
	public ObjectName getMBeanName () {
		return this.mBeanName;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.mib.SnmpMibExternalNode#getOid()
	 */
	public SnmpOid getOid () {
		return this.oid;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.mib.SnmpMibExternalNode#getSnmpDataType()
	 */
	public SnmpDataType getSnmpDataType () {
		return this.snmpDataType;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.mib.SnmpMibExternalNode#getJmxDataType()
	 */
	public Class<?> getJmxDataType () {
		return this.jmxDataType;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.mib.SnmpMibExternalNode#getValue()
	 */
	public Object getValue () throws Exception {
		return this.server.getAttribute(this.mBeanName, this.attributeName);
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.mib.SnmpMibExternalNode#setValue(java.lang.Object)
	 */
	public void setValue (final Object value) throws Exception {
		this.server.setAttribute(this.mBeanName, new Attribute(this.attributeName, value));
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.mib.SnmpMibExternalNode#isReadable()
	 */
	public boolean isReadable () {
		return this.readable;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.mib.SnmpMibExternalNode#isWritable()
	 */
	public boolean isWritable () {
		return this.writable;
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString () {
		Object value;
		try {
			value = getValue();
		}
		catch (final Throwable e) {
			value = "ERROR: " + (e.getMessage() != null ? e.getMessage() : e.getClass().getName());
		}
		return this.oid + ": (" + this.snmpDataType + ") " + value;
	}

}