package net.sf.snmpadaptor4j.daemon.opennms;

import java.math.BigInteger;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Iterator;
import java.util.Map.Entry;
import net.sf.snmpadaptor4j.daemon.SnmpDaemon;
import net.sf.snmpadaptor4j.daemon.SnmpException;
import net.sf.snmpadaptor4j.daemon.conf.SnmpConfiguration;
import net.sf.snmpadaptor4j.daemon.mib.SnmpDataType;
import net.sf.snmpadaptor4j.daemon.mib.SnmpMib;
import net.sf.snmpadaptor4j.daemon.mib.SnmpMibExternalNode;
import net.sf.snmpadaptor4j.daemon.mib.SnmpOid;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.opennms.protocols.snmp.SnmpAgentHandler;
import org.opennms.protocols.snmp.SnmpAgentSession;
import org.opennms.protocols.snmp.SnmpBadConversionException;
import org.opennms.protocols.snmp.SnmpCounter32;
import org.opennms.protocols.snmp.SnmpCounter64;
import org.opennms.protocols.snmp.SnmpEndOfMibView;
import org.opennms.protocols.snmp.SnmpGauge32;
import org.opennms.protocols.snmp.SnmpIPAddress;
import org.opennms.protocols.snmp.SnmpInt32;
import org.opennms.protocols.snmp.SnmpNull;
import org.opennms.protocols.snmp.SnmpObjectId;
import org.opennms.protocols.snmp.SnmpOctetString;
import org.opennms.protocols.snmp.SnmpOpaque;
import org.opennms.protocols.snmp.SnmpParameters;
import org.opennms.protocols.snmp.SnmpPduBulk;
import org.opennms.protocols.snmp.SnmpPduPacket;
import org.opennms.protocols.snmp.SnmpPduRequest;
import org.opennms.protocols.snmp.SnmpPeer;
import org.opennms.protocols.snmp.SnmpSMI;
import org.opennms.protocols.snmp.SnmpSyntax;
import org.opennms.protocols.snmp.SnmpTimeTicks;
import org.opennms.protocols.snmp.SnmpUInt32;
import org.opennms.protocols.snmp.SnmpVarBind;
import org.opennms.protocols.snmp.asn1.AsnEncodingException;

/**
 * SNMP daemon implementation for <b>joesnmp</b> API.
 * @author Jean-Philippe MINETTI
 */
public final class OpennmsSnmpDaemon
		implements SnmpDaemon, SnmpAgentHandler {

	/**
	 * Logger.
	 */
	private final Logger logger = Logger.getLogger(OpennmsSnmpDaemon.class);

	/**
	 * SNMP configuration settings.
	 */
	private final SnmpConfiguration configuration;

	/**
	 * SNMP <b>M</b>anagement <b>I</b>nformation <b>B</b>ase (MIB).
	 */
	private final SnmpMib snmpMib;

	/**
	 * SNMP session.
	 */
	private SnmpAgentSession snmpSession = null;

	/**
	 * Constructor.
	 * @param configuration SNMP configuration settings.
	 * @param snmpMib SNMP <b>M</b>anagement <b>I</b>nformation <b>B</b>ase (MIB).
	 */
	public OpennmsSnmpDaemon (final SnmpConfiguration configuration, final SnmpMib snmpMib) {
		super();
		this.configuration = configuration;
		this.snmpMib = snmpMib;
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.SnmpDaemon#start()
	 */
	public synchronized void start () throws Exception {
		if (this.snmpSession == null) {
			if (this.logger.isTraceEnabled()) {
				this.logger.trace("SNMP daemon starting...");
			}
			if (this.logger.isDebugEnabled()) {
				this.logger.debug("SNMP daemon will listen on " + this.configuration.getListenerAddress() + ":" + this.configuration.getListenerPort());
			}
			final SnmpPeer peer = new SnmpPeer(InetAddress.getByName(this.configuration.getListenerAddress()), this.configuration.getListenerPort().intValue());
			final SnmpParameters params = peer.getParameters();
			switch (this.configuration.getListenerSnmpVersion().intValue()) {
				case 1:
					params.setVersion(SnmpSMI.SNMPV1);
					break;
				default:
					params.setVersion(SnmpSMI.SNMPV2);
					break;
			}
			final String readCommunity = this.configuration.getListenerReadCommunity();
			params.setReadCommunity((readCommunity != null ? readCommunity : "public"));
			final String writeCommunity = this.configuration.getListenerWriteCommunity();
			params.setWriteCommunity((writeCommunity != null ? writeCommunity : "private"));
			this.snmpSession = new SnmpAgentSession(this, peer);
			this.logger.info("SNMP daemon listen on " + this.configuration.getListenerAddress() + ":" + this.configuration.getListenerPort());
			if (this.logger.isTraceEnabled()) {
				this.logger.trace("SNMP daemon started");
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.SnmpDaemon#stop()
	 */
	public synchronized void stop () {
		if (this.snmpSession != null) {
			if (this.logger.isTraceEnabled()) {
				this.logger.trace("SNMP daemon stopping...");
			}
			this.snmpSession.close();
			this.snmpSession = null;
			if (this.logger.isTraceEnabled()) {
				this.logger.trace("SNMP daemon stopped");
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * @see net.sf.snmpadaptor4j.daemon.SnmpDaemon#isStarted()
	 */
	public boolean isStarted () {
		return (this.snmpSession != null);
	}

	/*
	 * (non-Javadoc)
	 * @see org.opennms.protocols.snmp.SnmpAgentHandler#SnmpAgentSessionError(org.opennms.protocols.snmp .SnmpAgentSession, int, java.lang.Object)
	 */
	public void SnmpAgentSessionError (final SnmpAgentSession session, final int error, final Object ref) {
		throw new SnmpException(error);
	}

	/*
	 * (non-Javadoc)
	 * @see org.opennms.protocols.snmp.SnmpAgentHandler#snmpReceivedPdu(org.opennms.protocols.snmp. SnmpAgentSession, java.net.InetAddress, int,
	 * org.opennms.protocols.snmp.SnmpOctetString, org.opennms.protocols.snmp.SnmpPduPacket)
	 */
	public void snmpReceivedPdu (final SnmpAgentSession session, final InetAddress manager, final int port, final SnmpOctetString community, final SnmpPduPacket pdu) {
		if (this.logger.isTraceEnabled()) {
			this.logger.trace(pdu.getRequestId() + ": in treating by snmpReceivedPdu...");
		}
		if (pdu instanceof SnmpPduBulk) {
			final SnmpPduRequest response = new SnmpPduRequest(SnmpPduPacket.RESPONSE);
			response.setRequestId(pdu.getRequestId());
			doBulk(pdu.toVarBindArray(), (SnmpPduBulk) pdu, response);
			try {
				session.send(new SnmpPeer(manager, port), response);
			}
			catch (final AsnEncodingException e) {
				this.logger.error(pdu.getRequestId() + ": error " + SnmpPduPacket.ErrTooBig + " (response too big)");
				final SnmpPduRequest errorResponse = new SnmpPduRequest(SnmpPduPacket.RESPONSE);
				errorResponse.setRequestId(pdu.getRequestId());
				errorResponse.setErrorStatus(SnmpPduPacket.ErrTooBig);
				errorResponse.setErrorIndex(0);
				try {
					session.send(new SnmpPeer(manager, port), errorResponse);
				}
				catch (final Throwable e1) {
					this.logger.error(pdu.getRequestId() + ": an error has occurred when the error response sending", e1);
				}
			}
			catch (final Throwable e) {
				this.logger.error(pdu.getRequestId() + ": error " + SnmpPduPacket.ErrGenError + " (general error)", e);
				final SnmpPduRequest errorResponse = new SnmpPduRequest(SnmpPduPacket.RESPONSE);
				errorResponse.setRequestId(pdu.getRequestId());
				errorResponse.setErrorStatus(SnmpPduPacket.ErrGenError);
				errorResponse.setErrorIndex(0);
				try {
					session.send(new SnmpPeer(manager, port), errorResponse);
				}
				catch (final Throwable e1) {
					this.logger.error(pdu.getRequestId() + ": an error has occurred when the error response sending", e1);
				}
			}
		}
		else {
			this.logger.error(pdu.getRequestId() + ": request ignored (PDU " + pdu.getClass().getSimpleName() + " unhandled)");
		}
		if (this.logger.isTraceEnabled()) {
			this.logger.trace(pdu.getRequestId() + ": treated by snmpReceivedPdu");
		}
	}

	/*
	 * (non-Javadoc)
	 * @see org.opennms.protocols.snmp.SnmpAgentHandler#snmpReceivedGet(org.opennms.protocols.snmp. SnmpPduPacket, boolean)
	 */
	public SnmpPduRequest snmpReceivedGet (final SnmpPduPacket pdu, final boolean getNext) {
		if (this.logger.isTraceEnabled()) {
			this.logger.trace(pdu.getRequestId() + ": in treating by snmpReceivedGet...");
		}
		final SnmpPduRequest response = new SnmpPduRequest(SnmpPduPacket.RESPONSE);
		response.setRequestId(pdu.getRequestId());
		final SnmpVarBind[] binds = pdu.toVarBindArray();
		if (binds.length > 20) {
			this.logger.error(pdu.getRequestId() + ": error " + SnmpPduPacket.ErrTooBig + " (too OIDs - limited to 20)");
			response.setErrorStatus(SnmpPduPacket.ErrTooBig);
			response.setErrorIndex(0);
		}

		// SNMP v1
		else if (pdu instanceof SnmpPduRequest) {
			int resultIndex = 1;
			try {
				SnmpOid oid;
				SnmpMibExternalNode mibNode;
				for (final SnmpVarBind varBind : binds) {
					oid = SnmpOid.newInstance(varBind.getName().getIdentifiers());
					if (getNext) {
						mibNode = this.snmpMib.next(oid);
						if (mibNode == null) {
							response.addVarBind(new SnmpVarBind(varBind.getName()));
							throw new SnmpException("none OID found after " + oid, SnmpPduPacket.ErrNoSuchName, Level.DEBUG);
						}
						if (this.logger.isDebugEnabled()) {
							this.logger.debug(pdu.getRequestId() + ": " + oid + " --next--> " + mibNode);
						}
						response.addVarBind(newSnmpVarBind(mibNode));
					}
					else {
						mibNode = this.snmpMib.find(oid);
						if (mibNode == null) {
							response.addVarBind(new SnmpVarBind(varBind.getName()));
							throw new SnmpException("OID " + oid + " not found", SnmpPduPacket.ErrNoSuchName, Level.DEBUG);
						}
						if (this.logger.isDebugEnabled()) {
							this.logger.debug(pdu.getRequestId() + ": " + mibNode);
						}
						varBind.setValue(newSnmpValue(mibNode));
						response.addVarBind(varBind);
					}
					resultIndex++;
				}
			}
			catch (final SnmpException e) {
				this.logger.log(e.getLoggerLevel(), pdu.getRequestId() + ": " + e.getMessage(), e.getCause());
				response.setErrorStatus(e.getErrorStatus());
				response.setErrorIndex(resultIndex);
			}
			catch (final Throwable e) {
				this.logger.error(pdu.getRequestId() + ": error " + SnmpPduPacket.ErrGenError + " (general error)", e);
				response.setErrorStatus(SnmpPduPacket.ErrGenError);
				response.setErrorIndex(resultIndex);
			}
		}

		// SNMP v2
		else if (pdu instanceof SnmpPduBulk) {
			doBulk(binds, (SnmpPduBulk) pdu, response);
		}

		else {
			this.logger.error(pdu.getRequestId() + ": error " + SnmpPduPacket.ErrGenError + " (PDU " + pdu.getClass().getSimpleName() + " unhandled)");
			response.setErrorStatus(SnmpPduPacket.ErrGenError);
			response.setErrorIndex(0);
		}
		if (this.logger.isTraceEnabled()) {
			this.logger.trace(pdu.getRequestId() + ": treated by snmpReceivedGet");
		}
		return response;
	}

	/**
	 * Treats a <code>get-bulk</code> operation (SNMP v2).
	 * @param binds OID list to apply a <code>get-bulk</code>.
	 * @param bulk <code>get-bulk</code> operation.
	 * @param response Response to the request.
	 */
	private void doBulk (final SnmpVarBind binds[], final SnmpPduBulk bulk, final SnmpPduRequest response) {
		int resultIndex = 1;
		try {
			SnmpOid oid;
			SnmpMibExternalNode mibNode;
			Iterator<Entry<SnmpOid, SnmpMibExternalNode>> mibNodeEntryIterator;
			int i;
			for (final SnmpVarBind varBind : binds) {
				oid = SnmpOid.newInstance(varBind.getName().getIdentifiers());
				if (resultIndex <= bulk.getNonRepeaters()) {
					mibNode = this.snmpMib.next(oid);
					if (mibNode != null) {
						if (this.logger.isDebugEnabled()) {
							this.logger.debug(bulk.getRequestId() + ": " + oid + " --next--> " + mibNode);
						}
						response.addVarBind(newSnmpVarBind(mibNode));
					}
					else {
						if (this.logger.isDebugEnabled()) {
							this.logger.debug(bulk.getRequestId() + ": " + oid + " --next--> END OF MIB");
						}
						response.addVarBind(new SnmpVarBind(varBind.getName(), new SnmpEndOfMibView()));
					}
				}
				else {
					mibNodeEntryIterator = this.snmpMib.nextSet(oid).entrySet().iterator();
					if (mibNodeEntryIterator.hasNext()) {
						i = 0;
						while (mibNodeEntryIterator.hasNext() && (i < bulk.getMaxRepititions())) {
							mibNode = mibNodeEntryIterator.next().getValue();
							if (this.logger.isDebugEnabled()) {
								if (i == 0) {
									this.logger.debug(bulk.getRequestId() + ": " + oid + " --next--> " + mibNode);
								}
								else {
									this.logger.debug(bulk.getRequestId() + ": " + String.format("%" + oid.toString().length() + "s", "") + " --next--> " + mibNode);
								}
							}
							response.addVarBind(newSnmpVarBind(mibNode));
							i++;
						}
						if (i < bulk.getMaxRepititions()) {
							if (this.logger.isDebugEnabled()) {
								this.logger.debug(bulk.getRequestId() + ": " + oid + " --next--> .1.9: END OF MIB");
							}
							response.addVarBind(new SnmpVarBind(".1.9", new SnmpEndOfMibView()));
						}
					}
					else {
						if (this.logger.isDebugEnabled()) {
							this.logger.debug(bulk.getRequestId() + ": " + oid + " --next--> END OF MIB");
						}
						response.addVarBind(new SnmpVarBind(varBind.getName(), new SnmpEndOfMibView()));
					}
				}
				resultIndex++;
			}
		}
		catch (final SnmpException e) {
			this.logger.log(e.getLoggerLevel(), bulk.getRequestId() + ": " + e.getMessage(), e.getCause());
			response.setErrorStatus(e.getErrorStatus());
			response.setErrorIndex(resultIndex);
		}
		catch (final Throwable e) {
			this.logger.error(bulk.getRequestId() + ": error " + SnmpPduPacket.ErrGenError + " (general error)", e);
			response.setErrorStatus(SnmpPduPacket.ErrGenError);
			response.setErrorIndex(resultIndex);
		}
	}

	/*
	 * (non-Javadoc)
	 * @see org.opennms.protocols.snmp.SnmpAgentHandler#snmpReceivedSet(org.opennms.protocols.snmp. SnmpPduPacket)
	 */
	public SnmpPduRequest snmpReceivedSet (final SnmpPduPacket pdu) {
		if (this.logger.isTraceEnabled()) {
			this.logger.trace(pdu.getRequestId() + ": in treating by snmpReceivedSet...");
		}
		final SnmpPduRequest response = new SnmpPduRequest(SnmpPduPacket.RESPONSE);
		response.setRequestId(pdu.getRequestId());
		final SnmpVarBind[] binds = pdu.toVarBindArray();
		if (binds.length > 20) {
			this.logger.error(pdu.getRequestId() + ": error " + SnmpPduPacket.ErrTooBig + " (too OIDs - limited to 20)");
			response.setErrorStatus(SnmpPduPacket.ErrTooBig);
			response.setErrorIndex(0);
		}
		else {
			int resultIndex = 1;
			try {
				SnmpOid oid;
				SnmpSyntax newValue;
				SnmpMibExternalNode mibNode;
				for (final SnmpVarBind varBind : binds) {
					oid = SnmpOid.newInstance(varBind.getName().getIdentifiers());
					newValue = varBind.getValue();
					mibNode = this.snmpMib.find(oid);
					if (mibNode == null) {
						response.addVarBind(new SnmpVarBind(varBind.getName()));
						throw new SnmpException("OID " + oid + " not found", SnmpPduPacket.ErrNoSuchName, Level.DEBUG);
					}
					if (this.logger.isDebugEnabled()) {
						this.logger.debug(pdu.getRequestId() + ": old " + mibNode);
					}
					setValue(mibNode, newValue);
					response.addVarBind(newSnmpVarBind(mibNode));
					if (this.logger.isDebugEnabled()) {
						this.logger.debug(pdu.getRequestId() + ": new " + mibNode);
					}
					resultIndex++;
				}
			}
			catch (final SnmpException e) {
				this.logger.log(e.getLoggerLevel(), pdu.getRequestId() + ": " + e.getMessage(), e.getCause());
				response.setErrorStatus(e.getErrorStatus());
				response.setErrorIndex(resultIndex);
			}
			catch (final Throwable e) {
				this.logger.error(pdu.getRequestId() + ": error " + SnmpPduPacket.ErrGenError + " (general error)", e);
				response.setErrorStatus(SnmpPduPacket.ErrGenError);
				response.setErrorIndex(resultIndex);
			}
		}
		if (this.logger.isTraceEnabled()) {
			this.logger.trace(pdu.getRequestId() + ": treated by snmpReceivedSet");
		}
		return response;
	}

	/**
	 * Creates and returns a new {@link SnmpVarBind} by a {@link SnmpMibExternalNode}.
	 * @param mibNode {@link SnmpMibExternalNode} containing all the necessary data for create the {@link SnmpVarBind}.
	 * @return New {@link SnmpVarBind}.
	 */
	private SnmpVarBind newSnmpVarBind (final SnmpMibExternalNode mibNode) {
		return new SnmpVarBind(new SnmpObjectId(mibNode.getOid().getOid()), newSnmpValue(mibNode));
	}

	/**
	 * Creates and returns a new {@link SnmpSyntax} by a {@link SnmpMibExternalNode}.
	 * @param mibNode {@link SnmpMibExternalNode} containing all the necessary data for create the {@link SnmpSyntax}.
	 * @return New {@link SnmpSyntax}.
	 */
	private SnmpSyntax newSnmpValue (final SnmpMibExternalNode mibNode) {
		SnmpSyntax snmpValue;
		if (!mibNode.isReadable()) {
			throw new SnmpException("MBean attribute is not readable", SnmpPduPacket.ErrNoAccess);
		}
		Object value;
		try {
			value = mibNode.getValue();
		}
		catch (final Throwable e) {
			throw new SnmpException("Impossible to read the attribute due to error", SnmpPduPacket.ErrNoAccess, e);
		}
		if (value == null) {
			snmpValue = new SnmpNull();
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.integer32) {
			snmpValue = new SnmpInt32(toInteger32(value));
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.unsigned32) {
			snmpValue = new SnmpUInt32(toUnsigned32(value));
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.gauge32) {
			snmpValue = new SnmpGauge32(toUnsigned32(value));
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.counter32) {
			snmpValue = new SnmpCounter32(toUnsigned32(value));
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.counter64) {
			snmpValue = new SnmpCounter64(toUnsigned64(value));
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.timeTicks) {
			snmpValue = new SnmpTimeTicks(toTimeTicks(value));
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.octetString) {
			snmpValue = new SnmpOctetString(toOctetString(value));
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.ipAddress) {
			snmpValue = new SnmpIPAddress(toIpAddress(value));
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.objectIdentifier) {
			snmpValue = new SnmpObjectId(toObjectIdentifier(value));
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.opaque) {
			snmpValue = new SnmpOpaque(toOctetString(value));
		}
		else {
			throw new SnmpException(mibNode.getSnmpDataType() + " unhandled", SnmpPduPacket.ErrWrongType);
		}
		return snmpValue;
	}

	/**
	 * Sets a new value to {@link SnmpMibExternalNode}.
	 * @param mibNode {@link SnmpMibExternalNode} to update.
	 * @param newValue New value encapsulated to an object {@link SnmpSyntax}.
	 */
	private void setValue (final SnmpMibExternalNode mibNode, final SnmpSyntax newValue) {
		if (!mibNode.isReadable()) {
			throw new SnmpException("MBean attribute is not readable", SnmpPduPacket.ErrNoAccess);
		}
		if (!mibNode.isWritable()) {
			throw new SnmpException("MBean attribute is not writable", SnmpPduPacket.ErrNotWritable);
		}
		Object value;
		if (newValue instanceof SnmpNull) {
			value = null;
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.integer32) {
			if (!(newValue instanceof SnmpInt32)) {
				throw new SnmpException("New value is not an integer32", SnmpPduPacket.ErrBadValue);
			}
			value = fromInteger32(mibNode.getJmxDataType(), ((SnmpInt32) newValue).getValue());
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.unsigned32) {
			if (!(newValue instanceof SnmpUInt32)) {
				throw new SnmpException("New value is not an unsigned32", SnmpPduPacket.ErrBadValue);
			}
			value = fromUnsigned32(mibNode.getJmxDataType(), ((SnmpUInt32) newValue).getValue());
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.gauge32) {
			if (!(newValue instanceof SnmpGauge32)) {
				throw new SnmpException("New value is not a gauge32", SnmpPduPacket.ErrBadValue);
			}
			value = fromUnsigned32(mibNode.getJmxDataType(), ((SnmpGauge32) newValue).getValue());
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.counter32) {
			throw new SnmpException("counter32 are not writable", SnmpPduPacket.ErrNotWritable);
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.counter64) {
			throw new SnmpException("counter64 are not writable", SnmpPduPacket.ErrNotWritable);
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.timeTicks) {
			if (!(newValue instanceof SnmpTimeTicks)) {
				throw new SnmpException("New value is not a timeTicks", SnmpPduPacket.ErrBadValue);
			}
			value = fromTimeTicks(mibNode.getJmxDataType(), ((SnmpTimeTicks) newValue).getValue());
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.octetString) {
			if (!(newValue instanceof SnmpOctetString)) {
				throw new SnmpException("New value is not an octetString", SnmpPduPacket.ErrBadValue);
			}
			value = fromOctetString(mibNode.getJmxDataType(), ((SnmpOctetString) newValue).getString());
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.ipAddress) {
			if (!(newValue instanceof SnmpIPAddress)) {
				throw new SnmpException("New value is not an ipAddress", SnmpPduPacket.ErrBadValue);
			}
			try {
				value = fromIpAddress(mibNode.getJmxDataType(), ((SnmpIPAddress) newValue).convertToIpAddress());
			}
			catch (final SnmpBadConversionException e) {
				throw new SnmpException("IP address is malformed", SnmpPduPacket.ErrBadValue);
			}
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.objectIdentifier) {
			if (!(newValue instanceof SnmpObjectId)) {
				throw new SnmpException("New value is not an objectIdentifier", SnmpPduPacket.ErrBadValue);
			}
			value = fromObjectIdentifier(mibNode.getJmxDataType(), ((SnmpObjectId) newValue).getIdentifiers());
		}
		else if (mibNode.getSnmpDataType() == SnmpDataType.opaque) {
			if (!(newValue instanceof SnmpOpaque)) {
				throw new SnmpException("New value is not an opaque", SnmpPduPacket.ErrBadValue);
			}
			value = fromOctetString(mibNode.getJmxDataType(), ((SnmpOpaque) newValue).getString());
		}
		else {
			throw new SnmpException(mibNode.getSnmpDataType() + " unhandled", SnmpPduPacket.ErrWrongType);
		}
		try {
			mibNode.setValue(value);
		}
		catch (final Throwable e) {
			throw new SnmpException("Impossible to write in the attribute due to error", SnmpPduPacket.ErrNotWritable, e);
		}
	}

	/**
	 * Converts a value to signed 32-bit integer.
	 * @param value Value to convert.
	 * @return Value as signed 32-bit integer.
	 */
	private Integer toInteger32 (final Object value) {
		Integer result;
		if (value instanceof Integer) {
			result = (Integer) value;
		}
		else if (value instanceof Byte) {
			result = new Integer(((Byte) value).intValue());
		}
		else if (value instanceof Short) {
			result = new Integer(((Short) value).intValue());
		}
		else {
			throw new SnmpException(value.getClass() + " is inconsistent with integer32", SnmpPduPacket.ErrWrongType);
		}
		return result;
	}

	/**
	 * Converts a SNMP signed 32-bit integer to an object.
	 * @param type Object type to put in JMX attribute.
	 * @param value Value as signed 32-bit integer.
	 * @return Object to put in JMX attribute.
	 */
	private Object fromInteger32 (final Class<?> type, final int value) {
		Object result;
		if (Integer.class.equals(type) || int.class.equals(type)) {
			result = new Integer(value);
		}
		else if (Byte.class.equals(type) || byte.class.equals(type)) {
			if ((value < Byte.MIN_VALUE) || (value > Byte.MAX_VALUE)) {
				throw new SnmpException("the value must be between " + Byte.MIN_VALUE + " and " + Byte.MAX_VALUE, SnmpPduPacket.ErrBadValue);
			}
			result = new Byte((byte) value);
		}
		else if (Short.class.equals(type) || short.class.equals(type)) {
			if ((value < Short.MIN_VALUE) || (value > Short.MAX_VALUE)) {
				throw new SnmpException("the value must be between " + Short.MIN_VALUE + " and " + Short.MAX_VALUE, SnmpPduPacket.ErrBadValue);
			}
			result = new Short((short) value);
		}
		else {
			throw new SnmpException(type + " is inconsistent with integer32", SnmpPduPacket.ErrWrongType);
		}
		return result;
	}

	/**
	 * Converts a value to non-negative 32-bit integer.
	 * @param value Value to convert.
	 * @return Value as non-negative 32-bit integer.
	 */
	private Long toUnsigned32 (final Object value) {
		Long result;
		if (value instanceof Long) {
			result = (Long) value;
		}
		else if (value instanceof Integer) {
			result = new Long(((Integer) value).longValue());
		}
		else if (value instanceof Byte) {
			result = new Long(((Byte) value).longValue());
		}
		else if (value instanceof Short) {
			result = new Long(((Short) value).longValue());
		}
		else {
			throw new SnmpException(value.getClass() + " is inconsistent with unsigned32", SnmpPduPacket.ErrWrongType);
		}
		final long maxValue = 2 * (long) Integer.MAX_VALUE + 1;
		if ((result.longValue() < 0) || (result.longValue() > maxValue)) {
			throw new SnmpException("the value must be between 0 and " + maxValue, SnmpPduPacket.ErrWrongType);
		}
		return result;
	}

	/**
	 * Converts a SNMP non-negative 32-bit integer to an object.
	 * @param type Object type to put in JMX attribute.
	 * @param value Value as non-negative 32-bit integer.
	 * @return Object to put in JMX attribute.
	 */
	private Object fromUnsigned32 (final Class<?> type, final long value) {
		Object result;
		if (value < 0) {
			throw new SnmpException("the value must be positive", SnmpPduPacket.ErrBadValue);
		}
		if (Long.class.equals(type) || long.class.equals(type)) {
			final long maxValue = 2 * (long) Integer.MAX_VALUE + 1;
			if (value > maxValue) {
				throw new SnmpException("the value must be between 0 and " + maxValue, SnmpPduPacket.ErrBadValue);
			}
			result = new Long(value);
		}
		else if (Integer.class.equals(type) || int.class.equals(type)) {
			if (value > Integer.MAX_VALUE) {
				throw new SnmpException("the value must be between 0 and " + Integer.MAX_VALUE, SnmpPduPacket.ErrBadValue);
			}
			result = new Integer((int) value);
		}
		else if (Byte.class.equals(type) || byte.class.equals(type)) {
			if (value > Byte.MAX_VALUE) {
				throw new SnmpException("the value must be between 0 and " + Byte.MAX_VALUE, SnmpPduPacket.ErrBadValue);
			}
			result = new Byte((byte) value);
		}
		else if (Short.class.equals(type) || short.class.equals(type)) {
			if (value > Short.MAX_VALUE) {
				throw new SnmpException("the value must be between 0 and " + Short.MAX_VALUE, SnmpPduPacket.ErrBadValue);
			}
			result = new Short((short) value);
		}
		else {
			throw new SnmpException(type + " is inconsistent with unsigned32", SnmpPduPacket.ErrWrongType);
		}
		return result;
	}

	/**
	 * Converts a value to non-negative 64-bit integer.
	 * @param value Value to convert.
	 * @return Value as non-negative 64-bit integer.
	 */
	private BigInteger toUnsigned64 (final Object value) {
		BigInteger result;
		if (value instanceof BigInteger) {
			result = (BigInteger) value;
		}
		else if (value instanceof Long) {
			result = BigInteger.valueOf(((Long) value).longValue());
		}
		else {
			throw new SnmpException(value.getClass() + " is inconsistent with unsigned64", SnmpPduPacket.ErrWrongType);
		}
		final BigInteger minValue = BigInteger.valueOf(0);
		final BigInteger maxValue = BigInteger.valueOf(Long.MAX_VALUE).multiply(BigInteger.valueOf(2)).add(BigInteger.valueOf(1));
		if ((result.compareTo(minValue) < 0) || (result.compareTo(maxValue) > 0)) {
			throw new SnmpException("the value must be between " + minValue + " and " + maxValue, SnmpPduPacket.ErrWrongType);
		}
		return result;
	}

	/**
	 * Converts a value to non-negative 32-bit integer in hundredths of a second.
	 * @param value Value to convert in milliseconds.
	 * @return Value as non-negative 32-bit integer in hundredths of a second.
	 */
	private Long toTimeTicks (final Object value) {
		Long result;
		if (value instanceof Long) {
			result = new Long(((Long) value).longValue() / 10);
		}
		else {
			throw new SnmpException(value.getClass() + " is inconsistent with timeTicks", SnmpPduPacket.ErrWrongType);
		}
		final long maxValue = 2 * (long) Integer.MAX_VALUE + 1;
		if ((result.longValue() < 0) || (result.longValue() > maxValue)) {
			throw new SnmpException("the value must be between 0 and " + maxValue, SnmpPduPacket.ErrWrongType);
		}
		return result;
	}

	/**
	 * Converts a SNMP non-negative 32-bit integer in hundredths of a second to an object.
	 * @param type Object type to put in JMX attribute.
	 * @param value Value as non-negative 32-bit integer in hundredths of a second.
	 * @return Object to put in JMX attribute in milliseconds.
	 */
	private Object fromTimeTicks (final Class<?> type, final long value) {
		Object result;
		if (value < 0) {
			throw new SnmpException("the value must be positive", SnmpPduPacket.ErrBadValue);
		}
		if (Long.class.equals(type) || long.class.equals(type)) {
			final long maxValue = 2 * (long) Integer.MAX_VALUE + 1;
			if (value > maxValue) {
				throw new SnmpException("the value must be between 0 and " + maxValue, SnmpPduPacket.ErrBadValue);
			}
			result = new Long(10 * value);
		}
		else {
			throw new SnmpException(type + " is inconsistent with timeTicks", SnmpPduPacket.ErrWrongType);
		}
		return result;
	}

	/**
	 * Converts a value to array of ASCII code.
	 * @param value Value to convert.
	 * @return Value as array of ASCII code.
	 */
	private byte[] toOctetString (final Object value) {
		byte[] result;
		if (value instanceof String) {
			result = ((String) value).getBytes();
		}
		else if (value instanceof byte[]) {
			result = (byte[]) value;
		}
		else {
			result = value.toString().getBytes();
		}
		return result;
	}

	/**
	 * Converts a SNMP array of ASCII code to an object.
	 * @param type Object type to put in JMX attribute.
	 * @param value Value as ASCII code.
	 * @return Object to put in JMX attribute.
	 */
	private Object fromOctetString (final Class<?> type, final byte[] value) {
		Object result;
		if (String.class.equals(type)) {
			result = new String(value);
		}
		else if (byte[].class.equals(type)) {
			result = value;
		}
		else {
			throw new SnmpException(type + " is inconsistent with octetString", SnmpPduPacket.ErrWrongType);
		}
		return result;
	}

	/**
	 * Converts a value to internet address.
	 * @param value Value to convert.
	 * @return Value as internet address.
	 */
	private InetAddress toIpAddress (final Object value) {
		InetAddress result;
		if (value instanceof InetAddress) {
			result = (InetAddress) value;
		}
		else if (value instanceof String) {
			try {
				result = InetAddress.getByName((String) value);
			}
			catch (final UnknownHostException e) {
				throw new SnmpException("IP address is malformed", SnmpPduPacket.ErrWrongValue);
			}
		}
		else {
			throw new SnmpException(value.getClass() + " is inconsistent with ipAddress", SnmpPduPacket.ErrWrongType);
		}
		return result;
	}

	/**
	 * Converts a SNMP internet address to an object.
	 * @param type Object type to put in JMX attribute.
	 * @param value Value as internet address.
	 * @return Object to put in JMX attribute.
	 */
	private Object fromIpAddress (final Class<?> type, final InetAddress value) {
		Object result;
		if (InetAddress.class.equals(type)) {
			result = value;
		}
		else if (String.class.equals(type)) {
			result = value.getHostAddress();
		}
		else {
			throw new SnmpException(type + " is inconsistent with ipAddress", SnmpPduPacket.ErrWrongType);
		}
		return result;
	}

	/**
	 * Converts a value to OID of SNMP <b>M</b>anagement <b>I</b>nformation <b>B</b>ase (MIB).
	 * @param value Value to convert.
	 * @return Value as OID of SNMP <b>M</b>anagement <b>I</b>nformation <b>B</b>ase (MIB).
	 */
	private int[] toObjectIdentifier (final Object value) {
		int[] result;
		if (value instanceof SnmpOid) {
			result = ((SnmpOid) value).getOid();
		}
		else if (value instanceof String) {
			result = SnmpOid.newInstance((String) value).getOid();
		}
		else {
			throw new SnmpException(value.getClass() + " is inconsistent with objectIdentifier", SnmpPduPacket.ErrWrongType);
		}
		return result;
	}

	/**
	 * Converts an OID of SNMP <b>M</b>anagement <b>I</b>nformation <b>B</b>ase (MIB) to an object.
	 * @param type Object type to put in JMX attribute.
	 * @param value Value as OID of SNMP <b>M</b>anagement <b>I</b>nformation <b>B</b>ase (MIB).
	 * @return Object to put in JMX attribute.
	 */
	private Object fromObjectIdentifier (final Class<?> type, final int[] value) {
		Object result;
		if (SnmpOid.class.equals(type)) {
			result = SnmpOid.newInstance(value);
		}
		else if (String.class.equals(type)) {
			result = SnmpOid.newInstance(value).toString();
		}
		else {
			throw new SnmpException(type + " is inconsistent with objectIdentifier", SnmpPduPacket.ErrWrongType);
		}
		return result;
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString () {
		return "SnmpDaemon:opennms[configuration=" + this.configuration + "]";
	}

}